using System;
using System.IO;
using System.Text;
using System.Xml;
#if UNITY_EDITOR
using UnityEditor.Build.Reporting;
using UnityEngine.XR.OpenXR.Features.HisSupport;
#endif

namespace UnityEditor.XR.OpenXR.Features.HisSupport
{
#if UNITY_EDITOR
    internal class HisAndroidManifest : OpenXRFeatureBuildHooks
    {
        public override int callbackOrder => 1;
        public override Type featureType => typeof(HisOpenXRLoaderFeature);
        protected override void OnPreprocessBuildExt(BuildReport report) { }
        protected override void OnPostGenerateGradleAndroidProjectExt(string path)
        {
            var androidManifest = new AndroidManifest(GetManifestPath(path));
            androidManifest.AddHisElement();
            androidManifest.Save();
        }
        protected override void OnPostprocessBuildExt(BuildReport report) { }
        private string _manifestFilePath;
        private string GetManifestPath(string basePath)
        {
            if (!string.IsNullOrEmpty(_manifestFilePath)) return _manifestFilePath;
            var pathBuilder = new StringBuilder(basePath);
            pathBuilder.Append(Path.DirectorySeparatorChar).Append("src");
            pathBuilder.Append(Path.DirectorySeparatorChar).Append("main");
            pathBuilder.Append(Path.DirectorySeparatorChar).Append("AndroidManifest.xml");
            _manifestFilePath = pathBuilder.ToString();

            return _manifestFilePath;
        }
        private class AndroidXmlDocument : XmlDocument
        {
            private string m_Path;
            protected XmlNamespaceManager nsMgr;
            public readonly string AndroidXmlNamespace = "http://schemas.android.com/apk/res/android";

            public AndroidXmlDocument(string path)
            {
                m_Path = path;
                using (var reader = new XmlTextReader(m_Path))
                {
                    reader.Read();
                    Load(reader);
                }
                nsMgr = new XmlNamespaceManager(NameTable);
                nsMgr.AddNamespace("android", AndroidXmlNamespace);
            }
            public string Save()
            {
                return SaveAs(m_Path);
            }
            public string SaveAs(string path)
            {
                using (var writer = new XmlTextWriter(path, new UTF8Encoding(false)))
                {
                    writer.Formatting = Formatting.Indented;
                    Save(writer);
                }
                return path;
            }
        }
        private class AndroidManifest : AndroidXmlDocument
        {
            private readonly XmlElement ApplicationElement;
            private readonly XmlElement ManifestElement;

            public AndroidManifest(string path) : base(path)
            {
                ApplicationElement = SelectSingleNode("/manifest/application") as XmlElement;
                ManifestElement = SelectSingleNode("/manifest") as XmlElement;
            }
            private XmlAttribute CreateOrUpdateAndroidAttribute(string key, string value)
            {
                XmlAttribute attr = CreateAttribute("android", key, AndroidXmlNamespace);
                attr.Value = value;
                return attr;
            }

            private void CreateOrUpdateAndroidMetaData(string name, string value)
            {
                XmlNodeList nodeList = ApplicationElement.SelectNodes("meta-data");
                foreach (XmlNode node in nodeList)
                {
                    if (node != null)
                    {
                        foreach (XmlAttribute attribute in node.Attributes)
                        {
                            if (name.Equals(attribute.Value))
                            {
                                return;
                            }
                        }
                    }
                }

                var md = ApplicationElement.AppendChild(CreateElement("meta-data"));
                md.Attributes.Append(CreateOrUpdateAndroidAttribute("name", name.ToString()));
                md.Attributes.Append(CreateOrUpdateAndroidAttribute("value", value.ToString()));
            }

            private void CreateOrUpdateAndroidPermissions(string permission)
            {
                var xmlNodeList = ManifestElement.SelectNodes("uses-permission");

                foreach (XmlNode node in xmlNodeList)
                {
                    XmlAttribute nameAttr = (XmlAttribute)node.Attributes.GetNamedItem("name", AndroidXmlNamespace);
                    if (nameAttr != null && nameAttr.Value.Equals(permission))
                    {
                        return;
                    }
                }

                XmlElement newElement = CreateElement("uses-permission");
                newElement.SetAttribute("name", AndroidXmlNamespace, permission);
                ManifestElement.AppendChild(newElement);
            }

            private void CreateOrUpdateAndroidQueries(string tag, string key, string value)
            {
                var xmlNodeList = ManifestElement.SelectNodes("queries");

                XmlElement queriesElement = null;

                if (xmlNodeList != null && xmlNodeList.Count != 0)
                {
                    queriesElement = (XmlElement)(xmlNodeList[0]);

                    foreach (XmlNode node in xmlNodeList)
                    {
                        var list = node.SelectNodes(tag);

                        if (list == null || list.Count == 0)
                        {
                            continue;
                        }

                        foreach (XmlNode item in list)
                        {
                            XmlAttribute attr = (XmlAttribute)item.Attributes.GetNamedItem("name", AndroidXmlNamespace);
                            if (attr != null && attr.Value.Equals(value))
                            {
                                return;
                            }
                        }
                    }
                }

                if (queriesElement == null)
                {
                    queriesElement = CreateElement("queries");
                }

                XmlElement packageElement = CreateElement(tag);
                packageElement.SetAttribute(key, AndroidXmlNamespace, value);
                queriesElement.AppendChild(packageElement);
                ManifestElement.AppendChild(queriesElement);
            }

            private void CreateOrUpdateAndroidCategories(string key ,string activityCaterity)
            {
                var applicationNode = ManifestElement.SelectSingleNode("application");

                var xmlNodeList = applicationNode.SelectSingleNode("activity");

                var intentFilter = xmlNodeList.SelectSingleNode("intent-filter");

                var categories = intentFilter.SelectNodes("category");

                foreach (XmlNode node in categories)
                {
                    XmlAttribute nameAttr = (XmlAttribute)node.Attributes.GetNamedItem("name", AndroidXmlNamespace);
                    if (nameAttr != null && nameAttr.Value.Equals(activityCaterity))
                    {
                        return;
                    }
                }

                XmlElement packageElement = CreateElement("category");
                packageElement.SetAttribute(key, AndroidXmlNamespace, activityCaterity);
                intentFilter.AppendChild(packageElement);
            }

            internal void AddHisElement()
            {
                CreateOrUpdateAndroidMetaData("hxr_xr_app", "1");
                CreateOrUpdateAndroidPermissions("org.khronos.openxr.permission.OPENXR");
                //CreateOrUpdateAndroidPermissions("android.permission.WRITE_EXTERNAL_STORAGE");
                //CreateOrUpdateAndroidPermissions("android.permission.READ_EXTERNAL_STORAGE");
                CreateOrUpdateAndroidQueries("package", "name", "com.qualcomm.qti.openxrruntime");
                //CreateOrUpdateAndroidCategories("name", "com.qti.intent.category.SNAPDRAGON_VR");
            }
        }
    }
#endif
}

