﻿using System;
using System.Runtime.InteropServices;
using UnityEngine;

namespace UnityEngine.XR.OpenXR.Features.HisSupport
{
    public class HisHandDll
    {
        public const string LibFileName = "handtrackingplugin";

        public static readonly HxrXrVector3f HxrVector3fZero = new HxrXrVector3f() { x = 0, y = 0, z = 0 };
        public static readonly HxrXrQuaternionf HxrQuaternionZero = new HxrXrQuaternionf() { x = 0, y = 0, z = 0, w = 1 };

        public struct HxrHiHGInfoExt
        {
            public UInt32 frame_id;
            public ulong time_stamp;
            public Int32 hands;
            public float hand_scale;
            public HxrHiHGOneHand left;
            public HxrHiHGOneHand right;
        }

        [StructLayout(LayoutKind.Sequential), Serializable]
        public struct HxrHiHGOneHand
        {
            public int hand_id;
            public int left_right;
            public HxrXrPosef joint_wrist;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 32)]// 0-21
            public HxrXrPosef[] joints;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 5)]
            public HxrFingerProperties[] fingers;
            public HxrHiHGHandPoseParams theta;
            public HxrTransformRecognition trans_rec;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 21)]
            public HxrXrHandJointVelocityEXT[] joint_velocities;
        }

        [Serializable]
        public struct HxrXrPosef
        {
            public HxrXrQuaternionf orientation;
            public HxrXrVector3f position;
        }

        [StructLayout(LayoutKind.Sequential), Serializable]
        public struct HxrFingerProperties
        {
            public HxrFINGER_NOMINATE witch;
            public HxrFingerBendState curl;
            public HxrFingerBendState flexion;
            public HxrFingerAbductionState abduction;
            public HxrFingerOppositionState opposition;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 4)]
            public HxrXrPosef[] pose;
        };

        [StructLayout(LayoutKind.Sequential), Serializable]
        public struct HxrHiHGHandPoseParams
        {
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 26)]
            public float[] arrTheta;
        }

        [Serializable]
        public struct HxrXrHandJointVelocityEXT
        {
            public HxrXrVector3f linearVelocity;
            public HxrXrVector3f angularVelocity;
        };

        [Serializable]
        public struct HxrXrVector3f
        {
            public float x;
            public float y;
            public float z;
            public Vector3 ToVector3()
            {
                return new Vector3() { x = x, y = y, z = z };
            }
        };

        [Serializable]
        public struct HxrXrQuaternionf
        {
            public float x;
            public float y;
            public float z;
            public float w;

            public Quaternion ToQuat()
            {
                return new Quaternion() { x = x, y = y, z = z, w = w };
            }
        };


        [Serializable]
        public enum HxrFINGER_NOMINATE
        {
            HXR_HAND_THUMB = 0,
            HXR_HAND_INDEX = 1,
            HXR_HAND_MIDDLE = 2,
            HXR_HAND_RING = 3,
            HXR_HAND_LITTLE = 4
        };

        [Serializable]
        public enum HxrFingerBendState
        {
            HXR_FINGER_BEND_OPEN = 0,
            HXR_FINGER_BEND_NEUTRAL = 1,
            HXR_FINGER_BEND_CLOSE = 2,
        };

        [Serializable]
        public enum HxrFingerAbductionState
        {
            HXR_FINGER_ABDUCTION_CLOSED = 0,
            HXR_FINGER_ABDUCTION_OPEN = 1
        };

        [Serializable]
        public enum HxrFingerOppositionState
        {
            HXR_FINGER_OPPOSITION_NONE = 0,
            HXR_FINGER_OPPOSITION_NEAR = 1,
            HXR_FINGER_OPPOSITION_TOUCHING = 2,
        };

        [Serializable]
        public enum HxrTransformRecognition
        {
            HXR_TRANSFORM_WRIST_UP = 0,
            HXR_TRANSFORM_WRIST_DOWN = 1,
            HXR_TRANSFORM_PALM_UP = 2,
            HXR_TRANSFORM_PALM_DOWN = 3,
            HXR_TRANSFORM_PALM_TOWARDS_FACE = 4
        };

        public struct Posef
        {
            /// <summary>
            /// The orientation of hand joint.
            /// </summary>
            public HxrXrQuaternionf orientation;
            /// <summary>
            /// The position of hand joint.
            /// </summary>
            public HxrXrVector3f position;

            public UnityEngine.Quaternion ori;

            public UnityEngine.Vector3 pos;
            public UnityEngine.Vector3 dir;
            public override string ToString()
            {
                return string.Format("Orientation :{0}, {1}, {2}, {3}  Position: {4}, {5}, {6}",
                    orientation.x, orientation.y, orientation.z, orientation.w,
                    position.x, position.y, position.z);
            }

            public void ToHandPosef(HandType hand)
            {
                Vector3 pos = position.ToVector3();
                Quaternion rot = orientation.ToQuat();

                if (hand == HandType.Left)
                {
                    rot = new Quaternion(rot.x, rot.y, -rot.z, -rot.w) * new Quaternion(0.5f, -0.5f, 0.5f, -0.5f);
                }
                else
                {
                    rot = new Quaternion(rot.x, rot.y, -rot.z, -rot.w) * new Quaternion(-0.5f, -0.5f, -0.5f, -0.5f);
                }

                position.x = pos.x;
                position.y = pos.y;
                position.z = -pos.z;
                orientation.x = rot.x;
                orientation.y = rot.y;
                orientation.z = rot.z;
                orientation.w = rot.w;
            }

            public void ToJointPosef(HandType hand)
            {
                Vector3 pos = position.ToVector3();
                Quaternion rot = orientation.ToQuat();

                if (hand == HandType.Left)
                {
                    orientation.x = -rot.y;
                    orientation.y = rot.z;
                    orientation.z = rot.x;
                    orientation.w = -rot.w;
                }
                else
                {
                    orientation.x = rot.y;
                    orientation.y = -rot.z;
                    orientation.z = rot.x;
                    orientation.w = -rot.w;
                }

                position.x = pos.x;
                position.y = pos.y;
                position.z = -pos.z;
            }
        }
        public enum HandLocationStatus : ulong
        {
            /// <summary>
            /// Whether the joint's orientation is valid.
            /// </summary>
            OrientationValid = 0x00000001,
            /// <summary>
            /// Whether the joint's position is valid.
            /// </summary>
            PositionValid = 0x00000002,
            /// <summary>
            /// Whether the joint's orientation is being tracked.
            /// </summary>
            OrientationTracked = 0x00000004,
            /// <summary>
            /// Whether the joint's position is being tracked.
            /// </summary>
            PositionTracked = 0x00000008
        }

        public struct HandJointLocation
        {
            /// <summary>
            /// The status of hand joint location.
            /// </summary>
            public HandLocationStatus locationStatus;
            /// <summary>
            /// The orientation and position of hand joint.
            /// </summary>
            public Posef pose;
            /// <summary>
            /// The radius of hand joint.
            /// </summary>
            public float radius;
        }
        public struct HandJointLocations
        {
            /// <summary>
            /// The quality level of hand tracking:
            /// `0`: low
            /// `1`: high
            /// </summary>
            public uint isActive;
            /// <summary>
            /// The number of hand joints that the SDK supports. Currenty returns `26`.
            /// </summary>
            public uint jointCount;
            /// <summary>
            /// The scale of the hand.
            /// </summary>
            public float handScale;

            /// <summary>
            /// The locations (orientation and position) of hand joints.
            /// </summary>
            public HandJointLocation[] jointLocations;
        }

        public enum HandFinger
        {
            Thumb = 0,
            Index = 1,
            Middle = 2,
            Ring = 3,
            Pinky = 4
        }

        public enum HandJoint
        {
            JointWrist = 0,
            JointThumbMetacarpal = 1,
            JointThumbProximal = 2,
            JointThumbIntermediate = 3,
            JointThumbDistal = 4,
            JointThumbTip = 5,

            JointIndexProximal = 6,
            JointIndexIntermediate = 7,
            JointIndexDistal = 8,
            JointIndexTip = 9,

            JointMiddleProximal = 10,
            JointMiddleIntermediate = 11,
            JointMiddleDistal = 12,
            JointMiddleTip = 13,

            JointRingProximal = 14,
            JointRingIntermediate = 15,
            JointRingDistal = 16,
            JointRingTip = 17,

            JointLittleProximal = 18,
            JointLittleIntermediate = 19,
            JointLittleDistal = 20,
            JointLittleTip = 21,

            JointMax = 22
        }

        /// <summary>
        /// The data about the poses of ray and fingers.
        /// </summary>
        public struct HandAimState
        {
            /// <summary>
            /// The status of hand tracking. If it is not `tracked`, confidence will be `0`.
            /// </summary>
            public HandAimStatus aimStatus;
            /// <summary>
            /// The pose of the ray.
            /// </summary>
            public Posef aimRayPose;
            /// <summary>
            /// The strength of index finger's pinch.
            /// </summary>
            private float pinchStrengthIndex;
            /// <summary>
            /// The strength of middle finger's pinch.
            /// </summary>
            private float pinchStrengthMiddle;
            /// <summary>
            /// The strength of ring finger's pinch.
            /// </summary>
            private float pinchStrengthRing;
            /// <summary>
            /// The strength of little finger's pinch.
            /// </summary>
            private float pinchStrengthLittle;
            /// <summary>
            /// The strength of ray's touch.
            /// </summary>
            public float touchStrengthRay;
        }

        /// <summary>
        /// The status of ray and fingers.
        /// </summary>
        public enum HandAimStatus : ulong
        {
            /// <summary>
            /// Whether the data is valid.
            /// </summary>
            AimComputed = 0x00000001,
            /// <summary>
            /// Whether the ray appears.
            /// </summary>
            AimRayValid = 0x00000002,
            /// <summary>
            /// Whether the index finger pinches.
            /// </summary>
            AimIndexPinching = 0x00000004,
            /// <summary>
            /// Whether the middle finger pinches.
            /// </summary>
            AimMiddlePinching = 0x00000008,
            /// <summary>
            /// Whether the ring finger pinches.
            /// </summary>
            AimRingPinching = 0x00000010,
            /// <summary>
            /// Whether the little finger pinches.
            /// </summary>
            AimLittlePinching = 0x00000020,
            /// <summary>
            /// Whether the ray touches.
            /// </summary>
            AimRayTouched = 0x00000200
        }

        [DllImport(LibFileName)]
        public static extern void HxrHandTrackingStart();

        [DllImport(LibFileName)]
        public static extern void HxrHandTrackingStop();

        [DllImport(LibFileName)]
        public static extern void HxrHandTrackingGetData(IntPtr hxrHiInfo);

        [DllImport(LibFileName)]
        public static extern IntPtr HxrHandTrackingCommunication(int commType, string hxrMsg);

        [DllImport(LibFileName)]
        public static extern void HxrInitializeHandTracking();
    }
}